#! /bin/bash
# ----------------------------------------------------------------------------
# buildPackage.sh - Build the package on the current folder
#
# Jenkins environment variables must be set for proper functionality!
# Build job name must follow the form:
# <something>__<package-name>__<alpine-release>_<arch>
#
# Creation   :  2013-11-09  starwarsfan
#
# Copyright (c) 2013-2018 the eisfair team, team(at)eisfair(dot)org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# ----------------------------------------------------------------------------

#exec 2> /tmp/buildPackage-trace$$.txt
#set -x

# Backup where we came from
callDir=$(pwd)
ownLocation="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"
scriptName=$(basename $0)
cd ${ownLocation}
. ./include/helpers_log.sh
rtc=0



helpMe ()
{
    echo "

    Usage:
    ${0} [options]
        This script builds the package which is determined out of environment
        variable JOB_NAME and stores the result on the appropriate folder
        given by env var CI_RESULTFOLDER_EISFAIR_NG. These nvironment
        variables might be set using options described below. Otherwise the
        script will fail.
        Script should be executed out of repository root.

    Optional parameters:
    -a <alpine-version>
    -j <job-name>
        Set variable JOB_NAME. Value must follow the form
        'eisfair-ng/<alpine-release>/<stage>/<architecture>/<package>'
        so i. e. 'eisfair-ng/v3.1/testing/x86_64/bonnie'.
        Normally it is set by Jenkins as an env var.
    -r <result-folder>
        Set variable CI_RESULTFOLDER_EISFAIR_NG.
    -h  Show this help

    "
}



checkEnvironment ()
{
    info "##### Checking environment:"
    if [ -z "${JOB_NAME}" ] ; then
        helpMe
        die 100 "ERROR: Env var 'JOB_NAME' must be set or option -j must be used!"
    fi
    if [ -z "${CI_RESULTFOLDER_EISFAIR_NG}" ] ; then
        helpMe
        die 101 "ERROR: Env var 'CI_RESULTFOLDER_EISFAIR_NG' must be set or option -r must be used!"
    fi
    info "Done"
}



extractVariables ()
{
    # Extract package name, stage and arch from eisfair-ng/<alpine-release>/<stage>/<architecture>/<package>
    # Example: JOB_NAME='eisfair-ng/v3.1/testing/x86_64/bonnie'
    packageName=${JOB_NAME##*\/}
    packageArch=${JOB_NAME%\/*}
    packageArch=${packageArch#*\/}
    packageStage=${packageArch%\/*}
    packageStage=${packageStage#*\/}
    packageArch=${packageArch##*\/}
    if [ -z "${ALPINE_VERSION}" ] ; then
        alpineRelease=${packageStage%\/*}
    else
        alpineRelease=${ALPINE_VERSION}
    fi
}



buildPackage ()
{
    info "Updating pkg repository"
    sudo apk update

    info "Cd to ${packageName}"
    cd ../${packageName}

    info "Removing previously build apk files"
    rm -f *.apk

#    echo "Updating checksums"
#    abuild checksum

    info "Building package"
    abuild -r -c
    rtc=$?
    if [ "${rtc}" = 0 ] ; then
        info "Copying apk file(s) to repository folder"
        if ls *.apk >/dev/null 2>&1 ; then
            scp *.apk ${CI_RESULTFOLDER_EISFAIR_NG}/${alpineRelease}/${packageStage}/${packageArch}/
            rtc=$?
        else
            scp ~/packages/${packageName}/${packageArch}/*.apk ${CI_RESULTFOLDER_EISFAIR_NG}/${alpineRelease}/${packageStage}/${packageArch}/
            rtc=$?
        fi
    else
        die ${rtc} "Error during package build"
    fi
}

ALPINE_VERSION=''

while getopts a:j:r:h? option; do
    case ${option} in
        a) ALPINE_VERSION="${OPTARG}";;
        j) JOB_NAME="${OPTARG}";;
        r) CI_RESULTFOLDER_EISFAIR_NG="${OPTARG}";;
        h|?) helpMe && exit 0;;
        *) die 90 "invalid option \"${OPTARG}\"";;
    esac
done

checkEnvironment
extractVariables
buildPackage

exit ${rtc}

# ============================================================================
# End
# ============================================================================
