#! /bin/bash
# ----------------------------------------------------------------------------
# checkForNewPackages.sh - Check if every package on the repo${${jenkinsCliJar}
# jenkins job and if not, create one.
#
# Creation   :  2013-04-19  starwarsfan
#
# Copyright (c) 2013-2018 the eisfair team, team(at)eisfair(dot)org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# ----------------------------------------------------------------------------

#exec 2> /tmp/checkForNewPackages$$.log
#set -x

# Backup where we came from
callDir=$(pwd)
ownLocation="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"
scriptName=$(basename $0)
cd ${ownLocation}
. ./include/helpers_log.sh
rtc=0

# -----------------------------
# Check if settings file exists
if [ -f "settings.txt" ] ; then
    # ----------------------------------------------------------------
    # This should be the normal case: settings.txt exists, so load it.
    . settings.txt
    info "Settings loaded."
elif [ -f "settings.default.txt" ] ; then
    warning ""
    warning "Settings file not existing, creating new one out of default file."
    warning "The new one is on .gitignore, so it is secured that personal settings"
    warning "are not commited to the repository."
    cp settings.default.txt settings.txt
    warning "The script will be aborted here, please modify"
    warning "    ${ownLocation}/settings.txt"
    warning "to fit your needs. If the settings are OK, restart the script."
    warning ""
    die 2 ""
else
    # ---------------------------------------------------------------------
    # Thats odd: Not even 'settings.txt' nor 'settings.default.txt' exists!
    error ""
    error "ERROR: No settings file existing!"
    error "The default file 'settings.default.txt' must exist on the folder"
    error "'_ADMIN/'. This file is used to create"
    error "a personal settings file, which you can setup to your needs."
    error "Please check for that and restart the script again."
    error ""
    die 1 ""
fi



# ============================================================================
# The main part of the menu script
helpMe ()
{
    echo "

    Usage:  ${0##*/} [options]
        This script checks if a jenkins job for every package on this
        repository is existing. If not, jobs based on the templates will
        be created.

    Options:
    -n .. Do not trigger a build of new jobs immediately after their creation.
    -f .. Comma separated list of package folders following Jenkins logical structure e. g.
          'e-ng/previous/testing/x86_64,e-ng/previous/testing/x86'
    "
}


# ============================================================================
# See https://wiki.jenkins.io/display/JENKINS/Remote+access+API#RemoteaccessAPI-CSRFProtection
getCrumb() {
    info "Gettings current crumb"
    info "curl -u jenkins:************ '${jenkinsUrl}/crumbIssuer/api/xml?xpath=concat(//crumbRequestField,\":\",//crumb)'"
    jenkinsCrumb=$(curl -u "jenkins:${jenkinsPassword}" "${jenkinsUrl}/crumbIssuer/api/xml?xpath=concat(//crumbRequestField,\":\",//crumb)")
}



# ============================================================================
# Iterate over all folders on local working copy. In fact every folder
# represents a package (except folder _ADMIN), so for every folder the
# corresponding build jobs must exist or will be created.
iteratePackageFolders ()
{
    cd ${jobsFolder}

    local tmpCounter=0

	info "=============================================================================="
    for currentJobFolder in ${jobFolderList} ; do
        info ${currentJobFolder}
        local currentArch=${currentJobFolder##*/}

        info "Getting template job configuration"
        info "curl -u jenkins:************ -L -X GET ${jenkinsUrl}/job/${currentJobFolder}/job/${jobTemplateName}/config.xml -o ${tmpLocalTemplate}"
        curl -u jenkins:${jenkinsPassword} -L -X GET ${jenkinsUrl}/job/${currentJobFolder}/job/${jobTemplateName}/config.xml -o ${tmpLocalTemplate}

        info "Iterating over package folders"
        for currentFolder in $(ls -d ${workspaceFolder}/*/ | grep -v "_ADMIN") ; do
            info "${currentFolder}"
            # Cut last '/' and everything beyond
            currentCheckedPackage="${currentFolder%/*}"
            # Cut everything before last '/' including the '/' itself
            currentCheckedPackage=${currentCheckedPackage##*/}

            checkIfPackageArchMatches "${currentArch}" "${currentFolder}" "${currentCheckedPackage}"
            if [ $? = 0 ] ; then
#                info "createJob '$currentCheckedPackage' '$currentJobFolder' '${jobTemplateName}'"
                createJob "$currentCheckedPackage" "$currentJobFolder" ${jobTemplateName}
            fi
            # For testing: Only create 3 jobs
#            tmpCounter=$((tmpCounter+1))
#            if [ ${tmpCounter} -gt 3 ] ; then
#                break
#            fi
        done
        # For testing: Do only one step
#        break
    done
    rm -f ${tmpLocalTemplate} ${tmpUpdatedTemplate}
	info "=============================================================================="
}



# ============================================================================
# Check if the arch-entry on the current APKBUILD file matches the arch of the
# currently created build jobs.
#
# $1 .. Arch of current buildjobs
# $2 .. Folder of the current package for which a buildjob should be created
# $3 .. Name of the current package
#
# Return true if build job arch matches package arch, otherwise return false.
checkIfPackageArchMatches ()
{
    local _currentArch=$1
    local _currentFolder=$2
    local _currentPackage=$3
    if [ -f ${_currentFolder}/APKBUILD ] ; then
        local _currentPackageArch=$(grep "^arch=" ${_currentFolder}/APKBUILD | cut -d '=' -f2 | tr -cd '[[:alnum:]]')
        if [ -z "${_currentPackageArch}" ] ; then
            warning "Missing 'arch=...' on ${_currentPackage}/APKBUILD, skipping job creation"
            return 1
        elif [ "${_currentPackageArch}" = 'all' ] ; then
            return 0
        elif [ "${_currentPackageArch}" = "${_currentArch}" ] ; then
            return 0
        else
            info "Arch ${_currentPackageArch} of package ${_currentPackage} not matching current build arch ${_currentArch}, skipping job creation"
            return 1
        fi
    else
        info "Skipping package ${_currentPackage} as there is no APKBUILD file"
        return 1
    fi
}



# ============================================================================
# Create new jenkins job using Jenkins REST API
#
# $1 .. Package name
# $2 .. Logical Jenkins folder name, which should contain generated job
# $3 .. Name of the template-job which should be used
#       as the base for the new job
createJob ()
{
    local currentPackage=$1
    local logicalJobFolder=$2
    local jobTemplateName=$3
    local physicalJobFolder="$(echo "$logicalJobFolder" | sed "s#/job/#/jobs/#g")/jobs"
    local rtc=0

#    info "physicalJobFolder: ${physicalJobFolder}"

    if [ ! -d ${physicalJobFolder}/${currentPackage} -o ! -f ${physicalJobFolder}/${currentPackage}/config.xml ] ; then
        info "Creating job for package ${currentPackage}"
        # Update template for current package and activate it
        sed -e "s/TEMPLATE/$currentPackage/g" \
            -e "s#<disabled>true</disabled>#<disabled>false</disabled>#g" \
            ${tmpLocalTemplate} > ${tmpUpdatedTemplate}

        # Create job
        info "curl -u jenkins:************ -L -X POST ${jenkinsUrl}/job/${currentJobFolder}/createItem?name=${currentPackage} --data-binary  @${tmpUpdatedTemplate}  -H  ${jenkinsCrumb}  -H  Content-Type:text/xml"
        curl -u jenkins:${jenkinsPassword} -L -X POST ${jenkinsUrl}/job/${currentJobFolder}/createItem?name=${currentPackage} --data-binary "@${tmpUpdatedTemplate}" -H "${jenkinsCrumb}" -H "Content-Type:text/xml"
        rtc=$?
        if [ ${rtc} != 0 ] ; then
            error "ERROR: Something went wrong during creation of build-job '${logicalJobFolder}/${currentPackage}'"
        else
            # Trigger job
            info "curl -u jenkins:************ -L -X POST ${jenkinsUrl}/job/${currentJobFolder}/job/${currentPackage}/build -H  ${jenkinsCrumb}"
            curl -u jenkins:${jenkinsPassword} -L -X POST ${jenkinsUrl}/job/${currentJobFolder}/job/${currentPackage}/build -H "${jenkinsCrumb}"
            rtc=$?
            if [ ${rtc} != 0 ] ; then
                error "ERROR: Something went wrong during trigger of build-job '${logicalJobFolder}/${currentPackage}'"
            fi
        fi
    else
        info "Job for package ${currentPackage} already existing"
    fi
}



# Go into the repo root folder for the next steps and store it's full path
cd ${ownLocation}/..
workspaceFolder=$(pwd)

# Set some defaults
buildNewJobs=true
jobFolderList='job/e-ng/job/edge/job/testing/job/x86/job job/e-ng/job/edge/job/testing/job/x86_64/job'
tmpLocalTemplate=${workspaceFolder}/template.config.xml
tmpUpdatedTemplate=${workspaceFolder}/newPackage.config.xml

while getopts nf:h? option; do
    case ${option} in
        n) buildNewJobs=false;;
        f) jobFolderList=$(echo "${OPTARG}" | sed "s/,/ /g");;
        h|?) helpMe && exit 0;;
        *) die 90 "invalid option \"${OPTARG}\"";;
    esac
done

# $jobsFolder is defined on settings.txt
if [ ! -d ${jobsFolder} ] ; then
    die 100 "Jenkins job folder '$jobsFolder' not existing! Exiting."
fi

if [ ! -e "${jenkinsPasswordFile}" ] ; then
    die 101 "Password file not existing! Exiting."
fi
jenkinsPassword=$(cat ${jenkinsPasswordFile} )

getCrumb
iteratePackageFolders

exit ${rtc}

# ============================================================================
# End
# ============================================================================
