#! /bin/bash
# ----------------------------------------------------------------------------
# createRepoIndex.sh - Create the index for the repository
#
#
# Creation   :  2013-05-05  starwarsfan
#
# Copyright (c) 2013-2018 the eisfair team, team(at)eisfair(dot)org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# ----------------------------------------------------------------------------

#exec 2> /tmp/createRepoIndex-trace$$.txt
#set -x

# Backup where we came from
callDir=$(pwd)
ownLocation="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"
scriptName=$(basename $0)
cd ${ownLocation}
. ./include/helpers_log.sh
rtc=0

# -----------------------------
# Check if settings file exists
if [ -f "settings.txt" ] ; then
    # ----------------------------------------------------------------
    # This should be the normal case: settings.txt exists, so load it.
    . settings.txt
    info "Settings loaded."
elif [ -f "settings.default.txt" ] ; then
    warning ""
    warning "Settings file not existing, creating new one out of default file."
    warning "The new one is on .gitignore, so it is secured that personal settings"
    warning "are not commited to the repository."
    cp settings.default.txt settings.txt
    warning "The script will be aborted here, please modify"
    warning "    ${ownLocation}/settings.txt"
    warning "to fit your needs. If the settings are OK, restart the script."
    warning ""
    die 2 ""
else
    # ---------------------------------------------------------------------
    # Thats odd: Not even 'settings.txt' nor 'settings.default.txt' exists!
    error ""
    error "ERROR: No settings file existing!"
    error "The default file 'settings.default.txt' must exist on the folder"
    error "'_ADMIN/'. This file is used to create"
    error "a personal settings file, which you can setup to your needs."
    error "Please check for that and restart the script again."
    error ""
    die 1 ""
fi

helpMe ()
{
    echo "

    Usage:
    ${0} [parameters]
        This script creates the repository index file using the configured
        keypair and stores it on the configured path.

    Parameters:
    -v <version>
        .. The version of the system, for which the repository index should
           be created. Example: v3.7
    -a <architecture>
        .. The architecture of the system, for which the repository index
           should be created. Example: x86_64

    Optional parameters:
    -s <stage>
        .. The stage to be used on the repository. Default value: 'testing'
    -w <path-to-workspace>
        .. Path which shoul be used to work in.

    "
}

syncFromPkgRepo(){
    if [ ! -d "${pkgRepoDir}" ] ; then
        info "Creating work dir for all known packages"
        mkdir -p ${pkgRepoDir}
    fi
    info "Sync content from pkg repo"
    rsync --archive --delete -e ssh ${repoPath}/ ${pkgRepoDir}
    info "Sync'ed content:"
    ls -l ${pkgRepoDir}
}

# ============================================================================
# Create the repository index based on all determined packages and store it
# on pkg repository folder
createRepoIndex ()
{
    info "Creating repository index"

    if [ -d ${signingWorkDir} ] ; then
        info "Removing sign work dir"
        rm -rf ${signingWorkDir}
    fi
    mkdir -p ${signingWorkDir}
    cd ${signingWorkDir}

    # Setup version for the new index file
    CRTIMESTAMP=$(date +"%Y%m%d-%H%M%S")

    # See http://wiki.alpinelinux.org/wiki/Apkindex_format
    apk index -f -o APKINDEX.unsigned.tar.gz -d "v${CRTIMESTAMP}-$apkRepoQualifier" ${pkgRepoDir}/*.apk
    openssl dgst -sha1 -sign ~/.abuild/${signingPrivateKey} -out .SIGN.RSA.${signingPublicKey} APKINDEX.unsigned.tar.gz
    tar -c .SIGN.RSA.${signingPublicKey} | abuild-tar --cut | gzip -9 > signature.tar.gz
    cat signature.tar.gz APKINDEX.unsigned.tar.gz > ${pkgRepoDir}/APKINDEX.tar.gz

    # Cleanup
#    rm -f $signingWorkDir
}

uploadPkgIndex(){
    info "Uploading generated apk index"
    scp ${pkgRepoDir}/APKINDEX.tar.gz ${repoPath}/
}

# ============================================================================
# If the main repo was updated, create repo trigger files
createTriggerFiles ()
{
    if [ "$packageStage" = 'main' ] ; then
        info "Creating trigger file"
        touch ${apkRepositoryBaseFolder}/syncTrigger/${alpineRelease}__${packageArch}
    else
        info "Not on branch 'main', no pkg repo sync trigger files will be created"
    fi
}

alpineRelease=''
packageStage='testing'
packageArch=''
workspace="${ownLocation}"

while getopts a:s:v:w:h? option; do
    case ${option} in
        a) packageArch="${OPTARG}";;
        s) packageStage="${OPTARG}";;
        v) alpineRelease="${OPTARG}";;
        w) workspace="${OPTARG}";;
        h|?) helpMe && exit 0;;
        *) die 90 "invalid option \"${OPTARG}\"";;
    esac
done

[ -z "$alpineRelease" ] && die 91 "No version given!"
[ -z "$packageArch" ] && die 92 "No arch given!"
if [ "${workspace}" != "." ] ; then
    if [ ! -d "${workspace}" ] ; then
        mkdir ${workspace}
    fi
    cd ${workspace}
fi

signingWorkDir=${workspace}/repoSigning
pkgRepoDir=${workspace}/pkgRepo
repoPath=${CI_RESULTFOLDER_EISFAIR_NG}/${alpineRelease}/${packageStage}/${packageArch}

# Now do the job :-)
syncFromPkgRepo
createRepoIndex
uploadPkgIndex

#createTriggerFiles

exit ${rtc}

# ============================================================================
# End
# ============================================================================
