#! /bin/bash
# ----------------------------------------------------------------------------
# includePackagesIntoRepository.sh - Move manually uploaded apk's into the
#                                    corresponding repository folder and
#                                    update the repository index.
#
#
# Creation   :  2013-11-19  starwarsfan
#
# Copyright (c) 2013 the eisfair team, team(at)eisfair(dot)org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# ----------------------------------------------------------------------------

#exec 2> /tmp/createRepoIndex-trace$$.txt
#set -x

# Backup where we came from
callDir=$(pwd)
ownLocation="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"
scriptName=$(basename $0)
cd ${ownLocation}
. ./include/helpers_log.sh
rtc=0

# -----------------------------
# Check if settings file exists
if [ -f "settings.txt" ] ; then
    # ----------------------------------------------------------------
    # This should be the normal case: settings.txt exists, so load it.
    . settings.txt
    info "Settings loaded."
elif [ -f "settings.default.txt" ] ; then
    warning ""
    warning "Settings file not existing, creating new one out of default file."
    warning "The new one is on .gitignore, so it is secured that personal settings"
    warning "are not commited to the repository."
    cp settings.default.txt settings.txt
    warning "The script will be aborted here, please modify"
    warning "    ${ownLocation}/settings.txt"
    warning "to fit your needs. If the settings are OK, restart the script."
    warning ""
    die 2 ""
else
    # ---------------------------------------------------------------------
    # Thats odd: Not even 'settings.txt' nor 'settings.default.txt' exists!
    error ""
    error "ERROR: No settings file existing!"
    error "The default file 'settings.default.txt' must exist on the folder"
    error "'_ADMIN/'. This file is used to create"
    error "a personal settings file, which you can setup to your needs."
    error "Please check for that and restart the script again."
    error ""
    die 1 ""
fi



# ============================================================================
# Activate all uploaded packages for the given alpine version and architecture
# by moving them from the upload folder into the corresponding package
# repository folder.
activateUploadedPackages ()
{
    if [ -d "${sourcePath}" ] ; then
        # ToDo: Cleanup previous package versions

        if [ "$(ls -A ${sourcePath}/)" ] ; then
            info "Moving uploaded packages to repository folder"
            mv -f ${sourcePath}/* ${repoPath}/
            rtc=$?
            if [ ${rtc} != 0 ] ; then
                die ${rtc} " - Unable to move uploaded packages to repository folder!"
            fi

            info "Removing old packages"
            ./cleanupPackageFolder.sh -v ${alpineRelease} -b ${branch} -a ${alpineArch} -s ${amountOfPackagesToHold}
            rtc=$?
            if [ ${rtc} != 0 ] ; then
                warning " - Old packages could not be removed!"
            fi

            info "Updating package repository index"
            ./createRepoIndex.sh -v ${alpineRelease} -b ${branch} -a ${alpineArch}
            rtc=$?
            if [ ${rtc} != 0 ] ; then
                die ${rtc} " - Repository index could not be updated!"
            fi
        else
            info "No uploaded packages found on folder '$sourcePath'"
        fi
    else
        die 1 "Package source folder '$sourcePath' not existing!"
    fi
}



helpMe ()
{
    echo "

    Usage:
    ${0} -v <version> -a <architecture> [-b <branch>]
        Move manually uploaded packages into the corresponding repository
        folder and update the repository index. Additionally old package
        versions will be removed.

    Parameters:
    -v <version>
        .. The version of the system, for which all manually uploaded
           packages should be activated. Example: v2.7
    -a <architecture>
        .. The architecture of the system, for which all manually uploaded
           packages should be activated. Example: x86_64

    Optional parameters:
    -b <branch>
        .. The branch to be used on the repository. Default value: 'testing'
    -s <amount>
        .. Amount of versions per package which should be stay in place,
           all other (old) versions will be deleted. Default value: '3'

    "
}

alpineRelease=''
branch='testing'
alpineArch=''
amountOfPackagesToHold=3

while [ $# -ne 0 ]
do
    case $1 in
        -help|--help)
            # print usage
            usage
            exit 1
            ;;

        -v)
            if [ $# -ge 2 ] ; then
                alpineRelease=$2
                shift
            fi
            ;;

        -b)
            if [ $# -ge 2 ] ; then
                branch=$2
                shift
            fi
            ;;

        -a)
            if [ $# -ge 2 ] ; then
                alpineArch=$2
                shift
            fi
            ;;

        -s)
            if [ $# -ge 2 ] ; then
                amountOfPackagesToHold=$2
                shift
            fi
            ;;

        * )
            ;;
    esac
    shift
done

if [ -z "$alpineRelease" -o -z "$alpineArch" ] ; then
    die 1 "Parameters -v and -a must be used!"
fi

sourcePath=${apkManualUploadSourceFolder}/${alpineRelease}/${branch}/${alpineArch}
repoPath=${apkRepositoryBaseFolder}/${alpineRelease}/${branch}/${alpineArch}

# Now do the job :-)
activateUploadedPackages

# ============================================================================
# End
# ============================================================================
