#! /bin/bash
# ----------------------------------------------------------------------------
# releasePackage.sh - Release the package on the current folder
#
# Environment variables must be set for proper functionality!
#
# Creation   :  2013-11-13  starwarsfan
#
# Copyright (c) 2013-2018 the eisfair team, team(at)eisfair(dot)org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# ----------------------------------------------------------------------------

#exec 2> /tmp/releasePackage-trace$$.txt
#set -x

# Backup where we came from
callDir=$(pwd)
ownLocation="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"
scriptName=$(basename $0)
cd ${ownLocation}
. ./include/helpers_log.sh
rtc=0

branch='main'
amountOfPackagesToHold=3

helpMe ()
{
    echo "

    Usage:
    ${0} [options]
        This script releases the package which is given by the environment
        variable PACKAGE_TO_RELEASE or the corresponding cmdline option.
        Additionally the env var JOB_NAME must be set (or given via cmdline
        option) to determine base Alpine release and architecture. See
        description below. If both not given, the script will fail.
        Script should be executed out of repository root.

    Optional parameters:
    -p|--package-name <package-name>
        The package which should be released.
    -j|--job-name <job-name>
        Set variable JOB_NAME. Value must be set using the following schema:
        '<some-text>__<alpine-release>_<architecture>', e. g.:
        - releasePackage__v2.7_x86
        - releasePackage__v3.1_x86_64
    -b|--branch <branch>
        The branch to be used on the repository. Default value: 'main'
    -s <amount>
        Amount of versions per package which should be stay in place,
        all other (old) versions will be deleted. Default value: '3'

    "
}



checkEnvironment ()
{
    info "##### Checking environment:"
    if [ -z "${JOB_NAME}" ] ; then
        helpMe
        die 1 "ERROR: Env var 'JOB_NAME' must be set!"
    fi
    if [ -z "${PACKAGE_TO_RELEASE}" ] ; then
        helpMe
        die 2 "ERROR: Env var 'PACKAGE_TO_RELEASE' must be set!"
    fi
    if [ -z "${EISFAIR_NG_EDGE}" ] ; then
        helpMe
        die 3 "ERROR: Env var 'EISFAIR_NG_EDGE' must be set!"
    fi
    if [ -z "${EISFAIR_NG_STABLE}" ] ; then
        helpMe
        die 4 "ERROR: Env var 'EISFAIR_NG_STABLE' must be set!"
    fi
    if [ -z "${EISFAIR_NG_PREVIOUS}" ] ; then
        helpMe
        die 5 "ERROR: Env var 'EISFAIR_NG_PREVIOUS' must be set!"
    fi
    if [ -z "${CI_RESULTFOLDER_EISFAIR_NG}" ] ; then
        helpMe
        die 6 "ERROR: Env var 'CI_RESULTFOLDER_EISFAIR_NG' must be set!"
    fi
    echo "Done"
}



extractVariables ()
{
    # Extract package name from <some-text>__<alpine-release>_<arch>
    # Example:
    # eisfair-ng/releasePackage__edge_x86
    # eisfair-ng/releasePackage__edge_x86_64
    # eisfair-ng/releasePackage__stable_x86
    # eisfair-ng/releasePackage__stable_x86_64
    # eisfair-ng/releasePackage__previous_x86
    # eisfair-ng/releasePackage__previous_x86_64
    alpineRelease=$(echo ${JOB_NAME} | sed "s/\(.*__\)\([^_]*\)\(_\)\(.*\)/\2/g")
    packageArch=$(echo ${JOB_NAME}   | sed "s/\(.*__\)\([^_]*\)\(_\)\(.*\)/\4/g")
    # If stable or previous, replace with current corresponding version like v2.7 oder v2.6
    if [ "$alpineRelease" = 'stable' ] ; then
        alpineRelease=${EISFAIR_NG_STABLE}
    elif [ "$alpineRelease" = 'previous' ] ; then
        alpineRelease=${EISFAIR_NG_PREVIOUS}
    fi
    info "Building for Alpine Linux release '$alpineRelease' and architecture '$packageArch'"
}



releasePackage ()
{
    info "Updating pkg repository"
    sudo apk update

    info "Cd to ${PACKAGE_TO_RELEASE}"
    cd ../${PACKAGE_TO_RELEASE}

    info "Removing previously build apk files"
    rm -f *.apk

#    echo "Updating checksums"
#    abuild checksum

    info "Building package"
    abuild -r
    rtc=$?
    if [ "${rtc}" = 0 ] ; then
        info "Copying apk file(s) to repository folder"
        checkDeploymentDestination
        if ls *.apk >/dev/null 2>&1 ; then
            cp -f *.apk ${CI_RESULTFOLDER_EISFAIR_NG}/${alpineRelease}/${branch}/${packageArch}
            rtc=$?
        else
            cp -f ~/packages/${PACKAGE_TO_RELEASE}/${packageArch}/*.apk ${CI_RESULTFOLDER_EISFAIR_NG}/${alpineRelease}/${branch}/${packageArch}
            rtc=$?
        fi
        if [ "${rtc}" != 0 ] ; then
            die ${rtc} " - Unable to copy package to destination folder!"
        fi
    else
        die ${rtc} " - Package release build failed!"
    fi
    cd - > /dev/null
}



checkDeploymentDestination ()
{
    if [ ! -d ${CI_RESULTFOLDER_EISFAIR_NG}/${alpineRelease}/${branch}/${packageArch} ] ; then
        info "Repository directory not existing, creating it"
        mkdir -p ${CI_RESULTFOLDER_EISFAIR_NG}/${alpineRelease}/${branch}/${packageArch}
    fi
}



cleanupPackageFolder ()
{
    info "Removing old packages"
    ./cleanupPackageFolder.sh -v ${alpineRelease} -b ${branch} -a ${packageArch} -s ${amountOfPackagesToHold}
    rtc=$?
    if [ ${rtc} != 0 ] ; then
        warning " - Old packages could not be removed!"
    fi
}



updateRepoIndex ()
{
    # Example: -v v2.7 -b main -a x86_64
    info "Calling './createRepoIndex.sh -v ${alpineRelease} -b ${branch} -a ${packageArch}'"
    ./createRepoIndex.sh -v ${alpineRelease} -b ${branch} -a ${packageArch}
    rtc=$?
    if [ ${rtc} != 0 ] ; then
        die ${rtc} " - Repository index could not be updated!"
    fi
}



while [ $# -ne 0 ] ; do
    case $1 in
        -help|--help)
            # print usage
            usage
            exit 1
            ;;

        -j|--job-name)
            if [ $# -ge 2 ] ; then
                JOB_NAME=$2
                shift
            fi
            ;;

        -p|--package-name)
            if [ $# -ge 2 ] ; then
                PACKAGE_TO_RELEASE=$2
                shift
            fi
            ;;

        -b|--branch)
            if [ $# -ge 2 ] ; then
                branch=$2
                shift
            fi
            ;;

        -s)
            if [ $# -ge 2 ] ; then
                amountOfPackagesToHold=$2
                shift
            fi
            ;;

        * )
            ;;
    esac
    shift
done

checkEnvironment
extractVariables
releasePackage
cleanupPackageFolder
updateRepoIndex

exit ${rtc}

# ============================================================================
# End
# ============================================================================
