/*******************************************************************//**
 * @file server.c
 * (connection to database server)
 *
 * Copyright (C) 2006
 * Daniel Vogel, <daniel_vogel@t-online.de>
 *
 * Last Update:  $Id: server.h 28545 2011-06-13 19:27:26Z dv $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 **********************************************************************/

#include <stdlib.h>
#include <string.h>
#include "server.h"


static const char *PgServerGetEncoding(char *buffer, int buflen);

static SqlConnect_t *DefaultConnect = NULL;


/*******************************************************************//**
 * Establish connection to database server
 * @param host      host name to connect to or NULL
 * @param port      target port (as ascii number) or NULL
 * @param user      name of user to use for connection
 * @param passwd    password to use for connection or NULL
 * @param database  name of database to connect to
 * @return pointer to connection instance or NULL in case of failure
 **********************************************************************/

SqlConnect_t *PgServerConnect(
	const wchar_t *host, const wchar_t *port, 
	const wchar_t *user, const wchar_t *passwd,
	const wchar_t *database)
{
	SqlConnect_t *con = (SqlConnect_t*) malloc(sizeof(SqlConnect_t));
	if (con)
	{
		wchar_t connstr[256 + 1];
		char *mbstr;

		wcscpy(connstr, _T("dbname='"));
		wcscat(connstr, database);
		wcscat(connstr, _T("' user='"));
		wcscat(connstr, user);
		wcscat(connstr, _T("'"));
		if (passwd && wcslen(passwd))
		{
			wcscat(connstr, _T(" password='"));
			wcscat(connstr, passwd);
			wcscat(connstr, _T("'"));
		}
		if (host && wcslen(host))
		{
			wcscat(connstr, _T(" host='"));
			wcscat(connstr, host);
			wcscat(connstr, _T("'"));

			if (port && wcslen(port))
			{
				wcscat(connstr, _T(" port='"));
				wcscat(connstr, port);
				wcscat(connstr, _T("'"));
			}
		}

		mbstr = ModuleTCharToMbDup(connstr);
		if (mbstr)
		{
			con->HConnect = PQconnectdb(mbstr);
			free(mbstr);

			if (con->HConnect)
			{
				if (PgServerIsConnected(con))
				{
					char encoding[64 + 1];
				
					if (!DefaultConnect)
					{
						DefaultConnect = con;
					}

					PgServerGetEncoding(encoding, 64);

					PgExecSQL(con, _T("SET client_min_messages = 'ERROR';"));
					
					if ((strcasecmp(encoding, "utf-8") == 0) ||
					    (strcasecmp(encoding, "utf8")  == 0))
					{
						PgExecSQL(con, _T("SET client_encoding = 'UNICODE';"));
					}
					else
					{
						PgExecSQL(con, _T("SET client_encoding = 'LATIN9';"));
					}
				}
				con->User     = user     ? wcsdup(user) : NULL;
				con->Database = database ? wcsdup(database) : NULL;
				con->Password = passwd   ? wcsdup(passwd) : NULL;
				con->Host     = host     ? wcsdup(host) : NULL;
				con->Port     = port     ? wcsdup(port) : NULL;
				con->ErrMsg   = NULL;

				return con;
			}
		}
		free(con);
		return NULL;
	}
	return NULL;
}


/*******************************************************************//**
 * Check if database connection is established
 * @param  con      pointer to connection instance
 * @return true if connected, false otherwise
 **********************************************************************/

bool PgServerIsConnected(SqlConnect_t *con)
{
	return (con && (PQstatus(con->HConnect) == CONNECTION_OK));
}


/*******************************************************************//**
 * Close connection to database server
 * @param  con      pointer to connection instance
 * @return n/a
 **********************************************************************/

void PgServerDisconnect(SqlConnect_t *con)
{
	if (con)
	{
		if (DefaultConnect == con)
		{
			DefaultConnect = NULL;
		}
		PQfinish(con->HConnect);
		if (con->User)     free(con->User);
		if (con->Password) free(con->Password);
		if (con->Database) free(con->Database);
		if (con->Host)     free(con->Host);
		if (con->Port)     free(con->Port);
		if (con->ErrMsg)   free(con->ErrMsg);

		free(con);
	}
}


/*******************************************************************//**
 * Get last error message available on this connection
 * @param  con      pointer to connection instance
 * @return pointer to error message as wide character string
 **********************************************************************/

const wchar_t *PgServerGetError(SqlConnect_t *con)
{
	if (con)
	{
		if (con->ErrMsg)
		{
			free(con->ErrMsg);
		}
		con->ErrMsg = ModuleMbToTCharDup(PQerrorMessage(con->HConnect));
		return con->ErrMsg;
	}
	else
	{
		return _T("No error message available!");
	}
}


/*******************************************************************//**
 * Returns the password used to open the connection
 * @param  con      pointer to connection instance
 * @return password or empty string
 **********************************************************************/

const wchar_t *PgServerPasswd(SqlConnect_t *con)
{
	if (con && con->Password)
	{
		return con->Password;
	}
	return _T("");
}


/*******************************************************************//**
 * Returns the user used to open the connection
 * @param  con      pointer to connection instance
 * @return user name or empty string
 **********************************************************************/

const wchar_t *PgServerUser(SqlConnect_t *con)
{
	if (con && con->User)
	{
		return con->User;
	}
	return _T("");
}


/*******************************************************************//**
 * Returns the host used to open the connection
 * @param  con      pointer to connection instance
 * @return user name or empty string
 **********************************************************************/

const wchar_t *PgServerHost(SqlConnect_t *con)
{
	if (con && con->Host)
	{
		return con->Host;
	}
	return _T("");
}


/*******************************************************************//**
 * Returns the port used to open the connection
 * @param  con      pointer to connection instance
 * @return port number or empty string
 **********************************************************************/

const wchar_t *PgServerPort(SqlConnect_t *con)
{
	if (con && con->Port)
	{
		return con->Port;
	}
	return _T("");
}


/*******************************************************************//**
 * Opens a connection to the same server with a different database
 * target but with the same user authentication
 * @param  con       pointer to connection instance
 * @param  database  name of database to connect to
 * @return new connection instance or NULL in case of failure
 **********************************************************************/

SqlConnect_t *PgServerDupTo(SqlConnect_t *con, const wchar_t *database)
{
	if (con)
	{
		return PgServerConnect(
			con->Host, con->Port,
			con->User, con->Password,
			database);
	}
	return NULL;
}


/*******************************************************************//**
 * Returns the first database connection (used as default)
 * @param  n/a
 * @return connection instance or NULL in case of failure
 **********************************************************************/

SqlConnect_t *PgServerDefault(void)
{
	return DefaultConnect;
}


/*******************************************************************//**
 * Execute an SQL statement and return a result set, if the was statement
 * successfully executed by the server
 * @param  con       pointer to connection instance
 * @param  sqlstmt   SQL statement to execute
 * @return pointer to query result instance
 **********************************************************************/
 
SqlResult_t *PgQuerySQL(SqlConnect_t *con, const wchar_t *sqlstmt)
{
	if (con && con->HConnect)
	{
		char *mbstmt = ModuleTCharToMbDup(sqlstmt);
		if (mbstmt)
		{
			PGresult *res = PQexec(con->HConnect, mbstmt);

			free(mbstmt);

			if (res)
			{
				ExecStatusType status = PQresultStatus(res);
				if ((status == PGRES_COMMAND_OK)||(status == PGRES_TUPLES_OK))
				{
					SqlResult_t *result = (SqlResult_t *) malloc(sizeof(SqlResult_t));
					result->Result     = res;
					result->RowIndex   = -1;
					result->NumRows    = PQntuples(res);
					result->NumColumns = PQnfields(res);

					if ((status == PGRES_TUPLES_OK) && (result->NumColumns > 0))
					{
						int index;

						result->Columns   = (wchar_t**)
							malloc(result->NumColumns * sizeof(wchar_t*));
						result->RowData   = (wchar_t**)
							malloc(result->NumColumns * sizeof(wchar_t*));

						for (index = 0; index < result->NumColumns; index++)
						{
							result->Columns[index]  = ModuleMbToTCharDup(PQfname(res, index));
							result->RowData[index]  = NULL;
						}
					}
					else
					{
						result->NumColumns = 0;
						result->Columns = NULL;
						result->RowData = NULL;
					}

					return result;
				}
				PQclear(res);
			}
		}
	}
	return NULL;
}


/*******************************************************************//**
 * Execute an SQL statement directly without returning an result set
 * @param  con   pointer to connection instance
 * @return true if successful, false otherwise
 **********************************************************************/

bool PgExecSQL(SqlConnect_t *con, const wchar_t *sqlstmt)
{
	SqlResult_t *res = PgQuerySQL(con, sqlstmt);
	if (res)
	{
		PgResultFree(res);
		return true;
	}
	return false;
}


/*******************************************************************//**
 * Status of the last operation
 * @param  result   pointer to query result instance
 * @return one value from enumeration SqlExecStatus_t
 **********************************************************************/

SqlExecStatus_t PgResultStatus(SqlResult_t *result)
{
	if (result)
	{
		switch(PQresultStatus(result->Result))
		{
		case PGRES_COMMAND_OK:
			return SQL_COMMAND_OK;
		case PGRES_TUPLES_OK:
			return SQL_DATA_READY;
		default:
			return SQL_ERROR;
		}
	}
	return SQL_ERROR;
}


/*******************************************************************//**
 * Returns the number of rows contained in the result set
 * @param  result   pointer to query result instance
 **********************************************************************/

int PgResultNumRows(SqlResult_t *result)
{
	return result->NumRows;
}


/*******************************************************************//**
 * Returns the number of columns contained in the result set
 * @param  result   pointer to query result instance
 **********************************************************************/

int PgResultNumColumns(SqlResult_t *result)
{
	return result->NumColumns;
}


/*******************************************************************//**
 * Return the name of the specified column
 * @param  result   pointer to query result instance
 **********************************************************************/

const wchar_t *PgResultColumnName(SqlResult_t *result, int index)
{
	if ((index >= 0) && (index < result->NumColumns))
	{
		return (result->Columns[index]) ? result->Columns[index] : _T("");
	}
	return _T("");
}


/*******************************************************************//**
 * Field size in the specified column
 * @param  result   pointer to query result instance
 **********************************************************************/

int PgResultColumnSize(SqlResult_t *result, int index)
{
	return PQfsize(result->Result, index);
}


/*******************************************************************//**
 * Move the cursor to the next result tuple. Note that after a result
 * set is returned, the cursor has to be moved with ResultFirst to the
 * first row.
 * @param  result   pointer to query result instance
 * @return true of successfull, false if record not found
 **********************************************************************/

bool PgResultFirst(SqlResult_t *result)
{
	if (result->NumRows > 0)
	{
		result->RowIndex = 0;

		if (result->RowData)
		{
			int index;
			const char* data;

			for (index = 0; index < result->NumColumns; index++)
			{
				if (result->RowData[index] != NULL)
				{
					free(result->RowData[index]);
					result->RowData[index] = NULL;
				}

				data = PQgetvalue(result->Result, result->RowIndex, index);
				if (data)
				{
					result->RowData[index] = ModuleMbToTCharDup(data);
				}
			}
		}		
		return true;
	}
	return false;
}


/*******************************************************************//**
 * Move the cursor to the previous result tuple. 
 * @param  result   pointer to query result instance
 * @return true of successfull, false if record not found
 **********************************************************************/

bool PgResultPrevious(SqlResult_t *result)
{
	if (result->RowIndex > 0)
	{
		result->RowIndex--;

		if (result->RowData)
		{
			int index;
			const char* data;

			for (index = 0; index < result->NumColumns; index++)
			{
				if (result->RowData[index] != NULL)
				{
					free(result->RowData[index]);
					result->RowData[index] = NULL;
				}

				data = PQgetvalue(result->Result, result->RowIndex, index);
				if (data)
				{
					result->RowData[index] = ModuleMbToTCharDup(data);
				}
			}
		}		
		return true;
	}
	return false;
}


/*******************************************************************//**
 * Move the cursor to the next result tuple. 
 * @param  result   pointer to query result instance
 * @return true of successfull, false if record not found
 **********************************************************************/

bool PgResultNext(SqlResult_t *result)
{
	if (result->RowIndex < (result->NumRows - 1))
	{
		result->RowIndex++;

		if (result->RowData)
		{
			int index;
			const char* data;

			for (index = 0; index < result->NumColumns; index++)
			{
				if (result->RowData[index] != NULL)
				{
					free(result->RowData[index]);
					result->RowData[index] = NULL;
				}

				data = PQgetvalue(result->Result, result->RowIndex, index);
				if (data)
				{
					result->RowData[index] = ModuleMbToTCharDup(data);
				}
			}
		}
		return true;
	}
	return false;
}


/*******************************************************************//**
 * Move the cursor to the last result tuple.
 * @param  result   pointer to query result instance
 * @return true of successfull, false if record not found
 **********************************************************************/

bool PgResultLast(SqlResult_t *result)
{
	if (result->NumRows > 0)
	{
		result->RowIndex = (result->NumRows - 1);

		if (result->RowData)
		{
			int index;
			const char* data;

			for (index = 0; index < result->NumColumns; index++)
			{
				if (result->RowData[index] != NULL)
				{
					free(result->RowData[index]);
					result->RowData[index] = NULL;
				}

				data = PQgetvalue(result->Result, result->RowIndex, index);
				if (data)
				{
					result->RowData[index] = ModuleMbToTCharDup(data);
				}
			}
		}		
		return true;
	}
	return false;
}


/*******************************************************************//**
 * Read the data from the result set.
 * @param  result   pointer to query result instance
 * @param  index    field index within current record
 * @return cell data
 **********************************************************************/

const wchar_t *PgResultData(SqlResult_t *result, int index)
{
	if ((result->RowIndex >= 0) && (index >= 0) && (index < result->NumColumns))
	{
		return result->RowData[index];
	}
	else
	{
		return _T("");
	}
}


/*******************************************************************//**
 * is column 'index' a NULL value?
 * @param  result   pointer to query result instance
 * @para   index    field index within current record
 * @return true of field is NULL
 **********************************************************************/

bool PgResultIsNull(SqlResult_t *result, int index)
{
	return PQgetisnull(result->Result, result->RowIndex, index);
}


/*******************************************************************//**
 * Reset the data cursor. The next call to ResultFetch will address the
 * first row within the result set.
 * @param  result   pointer to query result instance
 * @return n/a
 **********************************************************************/

void PgResultReset(SqlResult_t *result)
{
	result->RowIndex = -1;
}


/*******************************************************************//**
 * Free the result set and all associated data
 * @param  result   pointer to query result instance
 * @return n/a
 **********************************************************************/

void PgResultFree(SqlResult_t *result)
{
	int i;

	if (result->Columns)
	{
		for (i = 0; i < result->NumColumns; i++)
		{
			if (result->Columns[i]) free(result->Columns[i]);
		}
	}
	if (result->RowData)
	{
		for (i = 0; i < result->NumColumns; i++)
		{
			if (result->RowData[i]) free(result->RowData[i]);
		}
	}
	if (result->Columns)   free(result->Columns);
	if (result->RowData)   free(result->RowData);
	PQclear(result->Result);
	free(result);
}


/*******************************************************************//**
 * Copy contents of environment variable $LC_CTYPE into a memory buffer
 * beginning with the char position right behind the fullstop
 * @param  buffer  buffer to store encoding to
 * @param  buflen  size of buffer
 * @return pointer to result buffer
 **********************************************************************/

static const char *PgServerGetEncoding(char *buffer, int buflen)
{
	const char *env = getenv("LC_CTYPE");
	
	buffer[0] = '\0';
	if (env)
	{
		env = strchr(env, '.');
		if (env)
		{
			strncpy(buffer, env + 1, buflen);
			buffer[buflen] = '\0';
		}
	}
	return buffer;
}
