/*******************************************************************//**
 * @file server.h
 * (connection to database server)
 *
 * Copyright (C) 2006
 * Daniel Vogel, <daniel_vogel@t-online.de>
 *
 * Last Update:  $Id: server.h 28545 2011-06-13 19:27:26Z dv $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 **********************************************************************/

#ifndef SERVER_H
#define SERVER_H

#include <stdbool.h>
#include "chartools.h"
#include <libpq-fe.h>


/*******************************************************************//**
 * @brief Connection instance type
 **********************************************************************/
 
typedef struct
{
	PGconn     *HConnect;
	wchar_t    *User;
	wchar_t    *Password;
	wchar_t    *Database;
	wchar_t    *Host;
	wchar_t    *Port;
	wchar_t    *ErrMsg;
} SqlConnect_t;


/*******************************************************************//**
 * @brief SQL query result type
 **********************************************************************/
 
typedef struct
{
	int         RowIndex;
	int         NumRows;
	int         NumColumns;
	PGresult   *Result;
	wchar_t   **Columns;
	wchar_t   **RowData;
} SqlResult_t;


/*******************************************************************//**
 * @brief SQL query status enumeration type
 **********************************************************************/

typedef enum
{
	SQL_COMMAND_OK = 0,
	SQL_DATA_READY = 1,
	SQL_ERROR      = 2,
} SqlExecStatus_t;


/***********************************************************************
 * Module interface implementation
 **********************************************************************/

SqlConnect_t   *PgServerConnect(const wchar_t *host, const wchar_t *port, 
                                const wchar_t *user, const wchar_t *passwd,
                                const wchar_t *database);
void            PgServerDisconnect(SqlConnect_t *con);
bool            PgServerIsConnected(SqlConnect_t *con);
const wchar_t  *PgServerGetError(SqlConnect_t *con);
const wchar_t  *PgServerPasswd(SqlConnect_t *con);
const wchar_t  *PgServerUser(SqlConnect_t *con);
const wchar_t  *PgServerHost(SqlConnect_t *con);
const wchar_t  *PgServerPort(SqlConnect_t *con);

SqlConnect_t   *PgServerDupTo(SqlConnect_t *con, const wchar_t *database);

SqlResult_t    *PgQuerySQL(SqlConnect_t *con, const wchar_t *sqlstmt);
bool            PgExecSQL(SqlConnect_t *con, const wchar_t *sqlstmt);

SqlConnect_t   *PgServerDefault(void);

SqlExecStatus_t PgResultStatus(SqlResult_t *result);
int             PgResultNumRows(SqlResult_t *result);
int             PgResultNumColumns(SqlResult_t *result);
const wchar_t  *PgResultColumnName(SqlResult_t *result, int index);
int             PgResultColumnSize(SqlResult_t *result, int index);
bool            PgResultFirst(SqlResult_t *result);
bool            PgResultPrevious(SqlResult_t *result);
bool            PgResultNext(SqlResult_t *result);
bool            PgResultLast(SqlResult_t *result);
const wchar_t  *PgResultData(SqlResult_t *result, int index);
bool            PgResultIsNull(SqlResult_t *result, int index);
void            PgResultReset(SqlResult_t *result);
void            PgResultFree(SqlResult_t *result);

#endif
