/*******************************************************************//**
 * @file server_api.c
 * (wrapper that acts as interface between cui and server.c)
 *
 * Copyright (C) 2006
 * Daniel Vogel, <daniel_vogel@t-online.de>
 *
 * Last Update:  $Id: server.h 28545 2011-06-13 19:27:26Z dv $
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 **********************************************************************/

#include <cui.h>
#include <cui-script.h>
#include <stdbool.h>
#include "chartools.h"
#include "server_api.h"

/***********************************************************************
 * global data
 **********************************************************************/

extern StartFrameProc    LibPqStartFrame;
extern InsertStrProc     LibPqInsertStr;
extern InsertIntProc     LibPqInsertInt;
extern InsertLongProc    LibPqInsertLong;
extern SendFrameProc     LibPqSendFrame;
extern ExecFrameProc     LibPqExecFrame;
extern WriteErrorProc    LibPqWriteError;

extern StubCreateProc    LibPqStubCreate;
extern StubCheckStubProc LibPqStubCheck;
extern StubDeleteProc    LibPqStubDelete;
extern StubSetHookProc   LibPqStubSetHook;
extern StubSetProcProc   LibPqStubSetProc;
extern StubFindProc      LibPqStubFind;


/***********************************************************************
 * local types
 **********************************************************************/

typedef struct DbConnection
{
	SqlConnect_t        *Connection;
	struct DbConnection *Next;
} DbConnection_t;

typedef struct DbResult
{
	SqlResult_t         *Result;
	struct DbResult     *Next;
} DbResult_t;


/***********************************************************************
 * local varaibles
 **********************************************************************/

static DbConnection_t *Connections = NULL;
static DbResult_t     *ResultSets  = NULL;


/***********************************************************************
 * local prototypes
 **********************************************************************/
 
static DbConnection_t *PgApiCreateConnection(SqlConnect_t* connection);
static bool            PgApiCheckConnection(SqlConnect_t* connection);
static bool            PgApiDeleteConnection(SqlConnect_t* connection);
static DbResult_t     *PgApiCreateResult(SqlResult_t* result);
static bool            PgApiCheckResult(SqlResult_t* result);
static bool            PgApiDeleteResult(SqlResult_t* result);
static wchar_t        *PgApiSingleLineDup(const wchar_t* data);


/*******************************************************************//**
 * Initialize API
 **********************************************************************/

void PgApiInit(void)
{
	Connections = NULL;
	ResultSets = NULL;
}

/*******************************************************************//**
 * Clear all open connections and result sets
 **********************************************************************/

void PgApiClear(void)
{
	DbConnection_t *conptr = Connections;
	DbResult_t     *resptr = ResultSets;

	while (resptr)
	{
		ResultSets = (DbResult_t*) resptr->Next;
		PgResultFree(resptr->Result);
		free(resptr);
		resptr = ResultSets;
	}

	while (conptr)
	{
		Connections = (DbConnection_t*) conptr->Next;
		PgServerDisconnect(conptr->Connection);
		free(conptr);
		conptr = Connections;
	}
}


/***********************************************************************
 * API functions
 **********************************************************************/

/*******************************************************************//**
 * Establish a server connection
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiServerConnect(int argc, const wchar_t* argv[])
{
	if (argc == 5)
	{
		SqlConnect_t* nc = PgServerConnect(
			argv[0], argv[1], argv[2], argv[3], argv[4]);
		if (nc)
		{
			PgApiCreateConnection(nc);
		}
		LibPqStartFrame(_T('R'), 48);
		LibPqInsertInt (ERROR_SUCCESS);
		LibPqInsertLong((unsigned long) nc);
		LibPqSendFrame ();
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Close a server connection
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiServerDisconnect(int argc, const wchar_t* argv[])
{
	if (argc == 1)
	{
		unsigned long connr;

		swscanf(argv[0], _T("%ld"), &connr);

		if (PgApiDeleteConnection((SqlConnect_t*)connr))
		{
			LibPqStartFrame(_T('R'), 32);
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Check if server connection is established
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiServerIsConnected(int argc, const wchar_t* argv[])
{
	if (argc == 1)
	{
		unsigned long connr;

		swscanf(argv[0], _T("%ld"), &connr);

		if (PgApiCheckConnection((SqlConnect_t*)connr))
		{
			LibPqStartFrame(_T('R'), 48);
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertInt (PgServerIsConnected((SqlConnect_t*)connr));
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Get error message
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiServerGetError(int argc, const wchar_t* argv[])
{
	if (argc == 1)
	{
		unsigned long connr;

		swscanf(argv[0], _T("%ld"), &connr);

		if (PgApiCheckConnection((SqlConnect_t*)connr))
		{
			const wchar_t* str = PgServerGetError((SqlConnect_t*)connr);

			LibPqStartFrame(_T('R'), 32 + wcslen(str) * sizeof(wchar_t));
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertStr (str);
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}

}


/*******************************************************************//**
 * Get password of this connection
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiServerPasswd(int argc, const wchar_t* argv[])
{
	if (argc == 1)
	{
		unsigned long connr;

		swscanf(argv[0], _T("%ld"), &connr);

		if (PgApiCheckConnection((SqlConnect_t*)connr))
		{
			const wchar_t* str = PgServerPasswd((SqlConnect_t*)connr);

			LibPqStartFrame(_T('R'), 32 + wcslen(str) * sizeof(wchar_t));
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertStr (str);
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Get user of this connection
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiServerUser(int argc, const wchar_t* argv[])
{
	if (argc == 1)
	{
		unsigned long connr;

		swscanf(argv[0], _T("%ld"), &connr);

		if (PgApiCheckConnection((SqlConnect_t*)connr))
		{
			const wchar_t* str = PgServerPasswd((SqlConnect_t*)connr);

			LibPqStartFrame(_T('R'), 32 + wcslen(str) * sizeof(wchar_t));
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertStr (str);
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Get host of this connection
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiServerHost(int argc, const wchar_t* argv[])
{
	if (argc == 1)
	{
		unsigned long connr;

		swscanf(argv[0], _T("%ld"), &connr);

		if (PgApiCheckConnection((SqlConnect_t*)connr))
		{
			const wchar_t* str = PgServerPasswd((SqlConnect_t*)connr);

			LibPqStartFrame(_T('R'), 32 + wcslen(str) * sizeof(wchar_t));
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertStr (str);
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Get port of this connection
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiServerPort(int argc, const wchar_t* argv[])
{
	if (argc == 1)
	{
		unsigned long connr;

		swscanf(argv[0], _T("%ld"), &connr);

		if (PgApiCheckConnection((SqlConnect_t*)connr))
		{
			const wchar_t* str = PgServerPort((SqlConnect_t*)connr);

			LibPqStartFrame(_T('R'), 32 + wcslen(str) * sizeof(wchar_t));
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertStr (str);
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Duplicate connection
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiServerDupTo(int argc, const wchar_t* argv[])
{
	if (argc == 2)
	{
		unsigned long connr;

		swscanf(argv[0], _T("%ld"), &connr);

		if (PgApiCheckConnection((SqlConnect_t*)connr))
		{
			SqlConnect_t* nc = PgServerDupTo((SqlConnect_t*)connr, argv[0]);
			if (nc)
			{
				PgApiCreateConnection(nc);
			}
			LibPqStartFrame(_T('R'), 48);
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertLong((unsigned long) nc);
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Execute a query
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiQuerySQL(int argc, const wchar_t* argv[])
{
	if (argc == 2)
	{
		unsigned long connr;

		swscanf(argv[0], _T("%ld"), &connr);

		if (PgApiCheckConnection((SqlConnect_t*)connr))
		{
			SqlResult_t* nr = PgQuerySQL((SqlConnect_t*)connr, argv[1]);
			if (nr)
			{
				PgApiCreateResult(nr);
			}
			LibPqStartFrame(_T('R'), 48);
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertLong((unsigned long) nr);
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Execute a SQL command
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiExecSQL(int argc, const wchar_t* argv[])
{
	if (argc == 2)
	{
		unsigned long connr;

		swscanf(argv[0], _T("%ld"), &connr);

		if (PgApiCheckConnection((SqlConnect_t*)connr))
		{
			LibPqStartFrame(_T('R'), 48);
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertInt (PgExecSQL((SqlConnect_t*)connr, argv[1]));
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Return default connection
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiServerDefault(int argc, const wchar_t* argv[])
{
	CUI_USE_ARG(argv);
	
	if (argc == 0)
	{
		SqlConnect_t* nc = PgServerDefault();

		LibPqStartFrame(_T('R'), 48);
		LibPqInsertInt (ERROR_SUCCESS);
		LibPqInsertLong((unsigned long) nc);
		LibPqSendFrame ();
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Return result status
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiResultStatus(int argc, const wchar_t* argv[])
{
	if (argc == 1)
	{
		unsigned long resnr;

		swscanf(argv[0], _T("%ld"), &resnr);

		if (PgApiCheckResult((SqlResult_t*)resnr))
		{
			LibPqStartFrame(_T('R'), 38);
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertInt ((int) PgResultStatus((SqlResult_t*)resnr));
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Return result status
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiResultNumRows(int argc, const wchar_t* argv[])
{
	if (argc == 1)
	{
		unsigned long resnr;

		swscanf(argv[0], _T("%ld"), &resnr);

		if (PgApiCheckResult((SqlResult_t*)resnr))
		{
			LibPqStartFrame(_T('R'), 32);
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertInt (PgResultNumRows((SqlResult_t*)resnr));
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Return number of columns
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiResultNumColumns(int argc, const wchar_t* argv[])
{
	if (argc == 1)
	{
		unsigned long resnr;

		swscanf(argv[0], _T("%ld"), &resnr);

		if (PgApiCheckResult((SqlResult_t*)resnr))
		{
			LibPqStartFrame(_T('R'), 32);
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertInt (PgResultNumColumns((SqlResult_t*)resnr));
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Return name of column
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiResultColumnName(int argc, const wchar_t* argv[])
{
	if (argc == 2)
	{
		unsigned long resnr;
		int index;

		swscanf(argv[0], _T("%ld"), &resnr);
		swscanf(argv[1], _T("%d"),  &index);

		if (PgApiCheckResult((SqlResult_t*)resnr))
		{
			const wchar_t* str = PgResultColumnName((SqlResult_t*)resnr, index);

			LibPqStartFrame(_T('R'), 32 + wcslen(str) * sizeof(wchar_t));
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertStr (str);
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Return size of column
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiResultColumnSize(int argc, const wchar_t* argv[])
{
	if (argc == 2)
	{
		unsigned long resnr;
		int           index;

		swscanf(argv[0], _T("%ld"), &resnr);
		swscanf(argv[1], _T("%d"), &index);

		if (PgApiCheckResult((SqlResult_t*)resnr))
		{
			LibPqStartFrame(_T('R'), 32);
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertInt (PgResultColumnSize((SqlResult_t*)resnr, index));
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Fetch next result line
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiResultFetch(int argc, const wchar_t* argv[])
{
	if (argc == 1)
	{
		unsigned long resnr;

		swscanf(argv[0], _T("%ld"), &resnr);

		if (PgApiCheckResult((SqlResult_t*)resnr))
		{
			LibPqStartFrame(_T('R'), 32);
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertInt (PgResultNext((SqlResult_t*)resnr));
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Fetch first result line
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiResultFirst(int argc, const wchar_t* argv[])
{
	if (argc == 1)
	{
		unsigned long resnr;

		swscanf(argv[0], _T("%ld"), &resnr);

		if (PgApiCheckResult((SqlResult_t*)resnr))
		{
			LibPqStartFrame(_T('R'), 32);
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertInt (PgResultFirst((SqlResult_t*)resnr));
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Fetch previous result line
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiResultPrevious(int argc, const wchar_t* argv[])
{
	if (argc == 1)
	{
		unsigned long resnr;

		swscanf(argv[0], _T("%ld"), &resnr);

		if (PgApiCheckResult((SqlResult_t*)resnr))
		{
			LibPqStartFrame(_T('R'), 32);
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertInt (PgResultPrevious((SqlResult_t*)resnr));
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Fetch last result line
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiResultLast(int argc, const wchar_t* argv[])
{
	if (argc == 1)
	{
		unsigned long resnr;

		swscanf(argv[0], _T("%ld"), &resnr);

		if (PgApiCheckResult((SqlResult_t*)resnr))
		{
			LibPqStartFrame(_T('R'), 32);
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertInt (PgResultLast((SqlResult_t*)resnr));
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Get column data
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiResultData(int argc, const wchar_t* argv[])
{
	if (argc == 2)
	{
		unsigned long resnr;
		int index;

		swscanf(argv[0], _T("%ld"), &resnr);
		swscanf(argv[1], _T("%d"), &index);

		if (PgApiCheckResult((SqlResult_t*)resnr))
		{
			const wchar_t* str = PgResultData((SqlResult_t*)resnr, index);
			
			LibPqStartFrame(_T('R'), 32 + wcslen(str) * sizeof(wchar_t));
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertStr (str);
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Check if result in column 'index' is NULL
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiResultIsNull(int argc, const wchar_t* argv[])
{
	if (argc == 2)
	{
		unsigned long resnr;
		int index;

		swscanf(argv[0], _T("%ld"), &resnr);
		swscanf(argv[1], _T("%d"), &index);

		if (PgApiCheckResult((SqlResult_t*)resnr))
		{
			LibPqStartFrame(_T('R'), 48);
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqInsertInt (PgResultIsNull((SqlResult_t*)resnr, index));
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Return cursor to the top of the result set
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiResultReset(int argc, const wchar_t* argv[])
{
	if (argc == 1)
	{
		unsigned long resnr;

		swscanf(argv[0], _T("%ld"), &resnr);

		if (PgApiCheckResult((SqlResult_t*)resnr))
		{
			PgResultReset((SqlResult_t*)resnr);

			LibPqStartFrame(_T('R'), 32);
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Free the result set
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiResultFree(int argc, const wchar_t* argv[])
{
	if (argc == 1)
	{
		unsigned long resnr;

		swscanf(argv[0], _T("%ld"), &resnr);

		if (PgApiDeleteResult((SqlResult_t*) resnr))
		{
			LibPqStartFrame(_T('R'), 32);
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/*******************************************************************//**
 * Copy result set to listview window
 * @param  argc  number of arguments
 * @param  argv  argument list
 * @return n/a
 **********************************************************************/

void PgApiResultToList(int argc, const wchar_t* argv[])
{
	if (argc >= 2)
	{
		WINDOWSTUB*   listview;
		SqlResult_t*    result;
		unsigned long tmplong;
		int           selindex = 0;
		int           keycolumn = -1;
		const wchar_t*  keyword = _T("");

		swscanf(argv[0], _T("%ld"), &tmplong);
		result = (SqlResult_t*) tmplong;
		
		swscanf(argv[1], _T("%ld"), &tmplong);
		listview = LibPqStubFind(tmplong);
		
		if (argc >= 3)
		{
			swscanf(argv[2], _T("%d"), &keycolumn);
		}
		if (argc >= 4)
		{
			keyword = argv[3];
		}
		
		if (listview && listview->Window && result)
		{
			int cols = PgResultNumColumns(result);
			int i;
			int index;
			
			while (PgResultNext(result))
			{
				LISTREC* rec = ListviewCreateRecord (listview->Window);
				if (rec)
				{
					int matched = FALSE;
					
					for (i = 0; i < cols; i++)
					{
						if (!PgResultIsNull(result, i))
						{
							wchar_t* data;
							
							data  = PgApiSingleLineDup(PgResultData(result, i));
							if (data)
							{
								ListviewSetColumnText(rec, i, data);
							
								if ((i == keycolumn) && (wcscmp(data, keyword) == 0))
								{
									matched = TRUE;
								}
								free(data);
							}
						}
					}
					index = ListviewInsertRecord(listview->Window, rec);
					if (matched)
					{
						selindex = index;
					}
				}
			}
			
			ListviewSetSel(listview->Window, selindex);
			
			LibPqStartFrame(_T('R'), 32);
			LibPqInsertInt (ERROR_SUCCESS);
			LibPqSendFrame ();
		}
		else
		{
			LibPqWriteError(ERROR_INVALID);
		}
	}
	else
	{
		LibPqWriteError(ERROR_ARGC);
	}
}


/***********************************************************************
 * Helper functions
 **********************************************************************/
 
/*******************************************************************//**
 * Create a connection structure
 * @param  connection  database connection instance
 * @return api wrapper of connection instance
 **********************************************************************/

static DbConnection_t *PgApiCreateConnection(SqlConnect_t *connection)
{
	DbConnection_t *newcon = (DbConnection_t*) malloc(sizeof(DbConnection_t));
	if (newcon)
	{
		newcon->Connection = connection;
		newcon->Next = Connections;
		Connections = newcon;
	}
	return newcon;
}


/*******************************************************************//**
 * Check if connection points to a valid connection
 * @param  connection  database connection instance
 * @return true if connection has been found, false otherwise
 **********************************************************************/

static bool PgApiCheckConnection(SqlConnect_t *connection)
{
	DbConnection_t *conptr = Connections;
	while (conptr)
	{
		if (conptr->Connection == connection)
		{
			return true;
		}
		conptr = conptr->Next;
	}
	return false;
}


/*******************************************************************//**
 * Free a connection if it exists
 * @param  connection  database connection instance
 * @return true if connection has been found, false otherwise
 **********************************************************************/

static bool PgApiDeleteConnection(SqlConnect_t *connection)
{
	DbConnection_t *conptr = Connections;
	DbConnection_t *oldptr = NULL;
	while (conptr)
	{
		if (conptr->Connection == connection)
		{
			PgServerDisconnect(conptr->Connection);
			if (oldptr)
			{
				oldptr->Next = conptr->Next;
			}
			else
			{
				Connections = conptr->Next;
			}
			free(conptr);
			return true;
		}
		oldptr = conptr;
		conptr = conptr->Next;
	}
	return false;
}


/*******************************************************************//**
 * Create a result structure
 * @param  result  database result instance
 * @return api wrapper of result instance
 **********************************************************************/

static DbResult_t *PgApiCreateResult(SqlResult_t *result)
{
	DbResult_t* newres = (DbResult_t*) malloc(sizeof(DbResult_t));
	if (newres)
	{
		newres->Result = result;
		newres->Next = ResultSets;
		ResultSets = newres;
	}
	return newres;
}


/*******************************************************************//**
 * Check if the result really exists
 * @param  result  database result instance
 * @return true if result object has been found, false otherwise
 **********************************************************************/

static bool PgApiCheckResult(SqlResult_t *result)
{
	DbResult_t *resptr = ResultSets;
	while (resptr)
	{
		if (resptr->Result == result)
		{
			return true;
		}
		resptr = (DbResult_t*) resptr->Next;
	}
	return false;
}


/*******************************************************************//**
 * Delete a result structure and free the result set
 * @param  result  database result instance
 * @return true if result object has been found, false otherwise
 **********************************************************************/

static bool PgApiDeleteResult(SqlResult_t *result)
{
	DbResult_t *resptr = ResultSets;
	DbResult_t *oldptr = NULL;
	while (resptr)
	{
		if (resptr->Result == result)
		{
			PgResultFree(resptr->Result);
			if (oldptr)
			{
				oldptr->Next = resptr->Next;
			}
			else
			{
				ResultSets = resptr->Next;
			}
			free(resptr);
			return true;
		}
		oldptr = resptr;
		resptr = (DbResult_t*) resptr->Next;
	}
	return false;
}


/*******************************************************************//**
 * the text in data is duplicated but characters like \n, \r and \t are
 * removed
 * @param  data  pointer to wide char string data
 * @result copy of data
 **********************************************************************/

static wchar_t *PgApiSingleLineDup(const wchar_t *data)
{
	wchar_t* line = (wchar_t*) malloc((wcslen(data) + 1) * sizeof(wchar_t));
	if (line)
	{
		wchar_t* l = line;
		while (*data != _T('\0'))
		{
			switch(*data)
			{
			case '\n':
			case '\t':
				*(l++) = _T(' ');
				break;
			case '\r':
				break;
			default:
				*(l++) = *data;
			}
			data++;
		}
		*l = _T('\0');
		return line;
	}
	return _T("");
}
